﻿using System;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using System.Windows.Threading;
using ELTE.Net.BallGame.Client.Model;

namespace ELTE.Net.BallGame.Client.View
{
    /// <summary>
    /// Főablak típusa.
    /// </summary>
    public partial class MainWindow : Window
    {
        private Thickness _BallStartPosition; // pozíciók, pontosabban határoló téglalapok
        private Thickness _BallCurrentPosition;
        private Thickness _BallNextPosition;
        private Thickness _PlayerPadStartPosition;
        private Thickness _PlayerPadCurrentPosition;
        private Thickness _OpponentPadStartPosition;
        private Thickness _OpponentPadCurrentPosition;
        private Int32 _GameStartCounter;
        private Boolean _PlayerStarts; // a játékos kezd-e
        private Boolean _Playing; // tart-e a játék
        private Double _Speed; // játék sebessége

        private ThicknessAnimation _BallAnimation; // animációk
        private ThicknessAnimation _PlayerPadAnimation;
        private ThicknessAnimation _OpponentPadAnimation;
        private DoubleAnimation _GameStartLabelFontSizeAnimation; // játék kezdési címke animálása
        private DoubleAnimation _GameStartLabelOpacityAnimation;

        private ConnectionWindow _ConnectionWindow;
        private GameBrowserWindow _GameBrowserWindow;
        private NetworkManager _Manager;

        private DispatcherTimer _Timer; // időzítő a játék kezdési címkéjéhez

        public MainWindow()
        {
            InitializeComponent();

            InitGui();
            InitGame();

            _Timer = new DispatcherTimer(); // időzítő felkonfigurálása
            _Timer.Interval = TimeSpan.FromSeconds(1);
            _Timer.Tick += new EventHandler(Timer_Tick);

            _Manager = new NetworkManager();
            _Manager.ConnectionChanged += new EventHandler(Manager_ConnectionChanged);
            _Manager.GamesReceived += new EventHandler<GamesEventArgs>(Manager_GamesReceived);
            _Manager.GameStepReceived += new EventHandler<GameStateEventArgs>(Manager_GameStepReceived);
            _Manager.GameStarted += new EventHandler(Manager_GameStarted);
            _Manager.GameOver += new EventHandler(Manager_GameOver);
        }

        protected override void OnClosing(System.ComponentModel.CancelEventArgs e)
        {
            _Manager.Disconnect();
            base.OnClosing(e);
        }
        /// <summary>
        /// Felület inicializálása.
        /// </summary>
        private void InitGui()
        {
            _EllipseBall.LayoutUpdated += new EventHandler(EllipseBall_LayoutUpdated);
            KeyDown += new KeyEventHandler(MainWindow_KeyDown);
            Closing += new System.ComponentModel.CancelEventHandler(MainWindow_Closing);

            _MenuConnectToGame.IsEnabled = true;
            _MenuDisconnectFromServer.IsEnabled = _MenuConnectToGame.IsEnabled = _MenuNewGame.IsEnabled = false;
            _LabelGameStart.Opacity = 0;
            _EllipseBall.Visibility = System.Windows.Visibility.Hidden;
            _RectangleOpponentPad.Visibility = System.Windows.Visibility.Hidden;
            _RectanglePlayerPad.Visibility = System.Windows.Visibility.Hidden;
        }
        /// <summary>
        /// Játék inicializálása.
        /// </summary>
        private void InitGame()
        {
            _BallStartPosition = new Thickness(_EllipseBall.Margin.Left, _EllipseBall.Margin.Top, 0, 0);
            _BallCurrentPosition = _BallStartPosition;
            _BallNextPosition = new Thickness();
            _PlayerPadStartPosition = _RectanglePlayerPad.Margin;
            _PlayerPadCurrentPosition = _PlayerPadStartPosition;
            _OpponentPadStartPosition = _RectangleOpponentPad.Margin;
            _OpponentPadCurrentPosition = _OpponentPadStartPosition;
            _Playing = false;
        }

        private void MenuExit_Click(object sender, RoutedEventArgs e)
        {
            Close();
        }

        private void Manager_GameStarted(object sender, EventArgs e)
        {
            Dispatcher.Invoke(new Action(() =>
            {
                _StatusBarText.Text = "Új játékos csatalkottt! Játék indul...";
                _GameStartCounter = 4;
                _Timer.Start();
            }));
        }

        private void Manager_GameOver(object sender, EventArgs e)
        {
            Dispatcher.Invoke(new Action(() =>
            {
                StopGame(true);
            }));
        }

        private void Manager_GameStepReceived(object sender, GameStateEventArgs e)
        {
            Dispatcher.Invoke(new Action(() =>
            {
                _OpponentPadCurrentPosition = _RectangleOpponentPad.Margin;
                AnimateOpponentPad(e.PadPosition);
            }));
        }

        private void Manager_GamesReceived(object sender, GamesEventArgs e)
        {
            Dispatcher.Invoke(new Action(() =>
            {
                if (_GameBrowserWindow != null)
                    _GameBrowserWindow.Games = e.Games;
            }));
        }

        private void Manager_ConnectionChanged(object sender, EventArgs e)
        {
            Dispatcher.Invoke(new Action(() =>
            {
                if (_Manager.Connected)
                {
                    _MenuConnectToServer.IsEnabled = false;
                    _MenuDisconnectFromServer.IsEnabled = _MenuConnectToGame.IsEnabled = _MenuNewGame.IsEnabled = true;
                    _StatusBarText.Text = "Kapcsolódva a szerverhez.";
                }
                else
                {
                    _MenuConnectToServer.IsEnabled = true;
                    _MenuDisconnectFromServer.IsEnabled = _MenuConnectToGame.IsEnabled = _MenuNewGame.IsEnabled = false;
                    _StatusBarText.Text = "Nincs kapcsolat a szerverrel.";
                }
            }));
        }

        private void MainWindow_Closing(object sender, System.ComponentModel.CancelEventArgs e)
        {
            _Manager.Disconnect();
        }

        private void MenuConnectToServer_Click(object sender, RoutedEventArgs e)
        {
            _ConnectionWindow = new ConnectionWindow();
            _ConnectionWindow.ShowDialog();
            if (_ConnectionWindow.DialogResult == true)
            {
                _Manager.Connect(_ConnectionWindow.ServerAddress, _ConnectionWindow.PlayerName);
            }
            _StatusBarText.Text = "Csatlakozás a szerverhez...";
        }

        private void MenuNewGame_Click(object sender, RoutedEventArgs e)
        {
            if (MessageBox.Show("Új játékot indítasz?", "Ping-Pong játék", MessageBoxButton.YesNo, MessageBoxImage.Question) == MessageBoxResult.Yes)
            {
                _PlayerStarts = true;
                _Manager.NewGame();
                _StatusBarText.Text = "Új játék elindítva. Várakozás másik játékra...";
            }
        }

        private void MenuDisconnectFromServer_Click(object sender, RoutedEventArgs e)
        {
            _Manager.Disconnect();
        }

        private void MenuConnectToGame_Click(object sender, RoutedEventArgs e)
        {
            _Manager.GetOpenGames();

            _GameBrowserWindow = new GameBrowserWindow();
            _GameBrowserWindow.ShowDialog();
            if (_ConnectionWindow.DialogResult == true)
            {
                _Manager.JoinGame(_GameBrowserWindow.SelectedGame);
                _PlayerStarts = false;
            }
        }
        /// <summary>
        /// Játék indítása.
        /// </summary>
        /// <param name="playerStarts">A játékos kezd-e.</param>
        private void StartGame(Boolean playerStarts)
        {
            _MenuNewGame.IsEnabled = false; // menükapcsolás
            _MenuConnectToGame.IsEnabled = false;
            _RectanglePlayerPad.Visibility = System.Windows.Visibility.Visible; // alaphelyzet beállítása
            _RectangleOpponentPad.Visibility = System.Windows.Visibility.Visible;
            _EllipseBall.Visibility = System.Windows.Visibility.Visible;

            _BallCurrentPosition = _BallStartPosition;
            _PlayerPadCurrentPosition = _PlayerPadStartPosition;
            _OpponentPadCurrentPosition = _OpponentPadStartPosition;

            _BallNextPosition.Left = _GameGrid.Width; // labda elindítása
            _BallNextPosition.Top = (playerStarts) ? 2 * _GameGrid.Height - _EllipseBall.Height : -_GameGrid.Height; // a kezdőirány attól függ, ki kezd

            _Speed = 1;
            _Playing = true;

            AnimateBall();
        }

        /// <summary>
        /// Játék leállítása.
        /// </summary>
        /// <param name="playerWon">A játékos győzött-e.</param>
        private void StopGame(Boolean playerWon)
        {
            if (_Playing) // csak ha fut a játék
            {
                if (_BallAnimation != null) _BallAnimation.Freeze(); // animációk leállítása
                if (_OpponentPadAnimation != null) _OpponentPadAnimation.Freeze();
                if (_PlayerPadAnimation != null) _PlayerPadAnimation.Freeze();
                _Playing = false;

                if (playerWon)
                {
                    MessageBox.Show("Gratulálok, nyertél! :)", "Ping-Pong játék", MessageBoxButton.OK, MessageBoxImage.Information);
                }
                else
                {
                    _Manager.EndGame(); // jelezzük a játék végét a szervernek
                    MessageBox.Show("Sajnos vesztettél! :(", "Ping-Pong játék", MessageBoxButton.OK, MessageBoxImage.Information);
                }

                _MenuNewGame.IsEnabled = true; // menükapcsolás
                _MenuConnectToGame.IsEnabled = true;
                _RectanglePlayerPad.Visibility = System.Windows.Visibility.Hidden;
                _RectangleOpponentPad.Visibility = System.Windows.Visibility.Hidden;
                _EllipseBall.Visibility = System.Windows.Visibility.Hidden;
            }
        }

        private void MainWindow_KeyDown(object sender, KeyEventArgs e)
        {
            if (_Playing) // billentyűzet lekezelése
            {
                _PlayerPadCurrentPosition = _RectanglePlayerPad.Margin;
                Int32 movement = 0;
                if (e.Key == Key.Left)
                    if (_RectanglePlayerPad.Margin.Left > 0)
                        movement = -100;
                if (e.Key == Key.Right)
                    if (_RectanglePlayerPad.Margin.Right > 0)
                        movement = 100;

                AnimatePlayerPad(movement);

                _Manager.StepGame(movement);
            }
        }

        private void Timer_Tick(object sender, EventArgs e) // időzített esemény
        {
            if (_GameStartCounter == 4) // kezdő felirat váltása a hátralévő idő függvényében
            {
                _LabelGameStart.Content = "Játék indul...";
                AnimateGameStartLabel();
                _GameStartCounter--;
            }
            else if (_GameStartCounter > 0)
            {
                _LabelGameStart.Content = _GameStartCounter.ToString();
                AnimateGameStartLabel();
                _GameStartCounter--;
            }
            else
            {
                _Timer.Stop(); // időzítő leállítása
                _StatusBarText.Text = "Játék elindult.";
                StartGame(_PlayerStarts); // játék indítása
            }
        }

        private void EllipseBall_LayoutUpdated(object sender, EventArgs e)
        {
            if (_Playing) // csak akkor, ha tart a játék
            {
                if (_EllipseBall.Margin.Top <= 0) // ha elértük a tetőpontot
                {
                    StopGame(true);
                    return;
                }
                _BallCurrentPosition = _EllipseBall.Margin;

                // játékos ütőjének ütközik balról:
                if ((_EllipseBall.Margin.Top + _EllipseBall.Height >= _RectanglePlayerPad.Margin.Top) && _EllipseBall.Margin.Left >= (_RectanglePlayerPad.Margin.Left - 30) && _EllipseBall.Margin.Left <= (_RectanglePlayerPad.Margin.Left))
                {
                    _Speed *= 1.05; // sebességnövelés
                    _BallNextPosition.Top = -_EllipseBall.Height;
                    _BallNextPosition.Left = _BallCurrentPosition.Left - 200;
                    AnimateBall();
                }
                // játékos ütőjének ütközik középen:
                if ((_EllipseBall.Margin.Top + _EllipseBall.Height >= _RectanglePlayerPad.Margin.Top) && _EllipseBall.Margin.Left > (_RectanglePlayerPad.Margin.Left) && _EllipseBall.Margin.Left < (_RectanglePlayerPad.Margin.Left + 80))
                {
                    _Speed *= 1.05; // sebességnövelés
                    _BallNextPosition.Top = -_EllipseBall.Height;
                    AnimateBall();
                }
                // játékos ütőjének ütközik jobbról: 
                else if ((_EllipseBall.Margin.Top + _EllipseBall.Height >= _RectanglePlayerPad.Margin.Top) && _EllipseBall.Margin.Left >= (_RectanglePlayerPad.Margin.Left + 80) && _EllipseBall.Margin.Left <= (_RectanglePlayerPad.Margin.Left + 110))
                {
                    _Speed *= 1.05;
                    _BallNextPosition.Top = -_EllipseBall.Height;
                    _BallNextPosition.Left = _BallCurrentPosition.Left + 200;
                    AnimateBall();
                }
                // ellenfél ütőjének ütközik balról:
                if ((_EllipseBall.Margin.Top <= _RectangleOpponentPad.Margin.Top + _RectangleOpponentPad.Height) && _EllipseBall.Margin.Left >= (_RectangleOpponentPad.Margin.Left - 30) && _EllipseBall.Margin.Left <= (_RectangleOpponentPad.Margin.Left))
                {
                    _Speed *= 1.05;
                    _BallNextPosition.Top = _GameGrid.Height;
                    _BallNextPosition.Left = _BallCurrentPosition.Left - 200;
                    AnimateBall();
                }
                // ellenfél ütőjének ütközik középen:
                if ((_EllipseBall.Margin.Top <= _RectangleOpponentPad.Margin.Top + _RectangleOpponentPad.Height) && _EllipseBall.Margin.Left > (_RectangleOpponentPad.Margin.Left) && _EllipseBall.Margin.Left < (_RectangleOpponentPad.Margin.Left + 80))
                {
                    _Speed *= 1.05;
                    _BallNextPosition.Top = _GameGrid.Height;
                    AnimateBall();
                }
                // ellenfél ütőjének ütközik jobbról: 
                else if ((_EllipseBall.Margin.Top <= _RectangleOpponentPad.Margin.Top + _RectangleOpponentPad.Height) && _EllipseBall.Margin.Left >= (_RectangleOpponentPad.Margin.Left + 80) && _EllipseBall.Margin.Left <= (_RectangleOpponentPad.Margin.Left + 110))
                {
                    _Speed *= 1.05;
                    _BallNextPosition.Top = _GameGrid.Height;
                    _BallNextPosition.Left = _BallCurrentPosition.Left + 200;
                    AnimateBall();
                }
                else if (_EllipseBall.Margin.Left <= 0) // bal falnak ütközik
                {
                    _BallNextPosition.Left = _GameGrid.Width;
                    AnimateBall();
                }
                else if (_EllipseBall.Margin.Left >= _GameGrid.Width - _EllipseBall.Width) // jobb falnak ütközik
                {
                    _BallNextPosition.Left = 0;
                    AnimateBall();
                }
                else if (_EllipseBall.Margin.Top + _EllipseBall.Height >= _GameGrid.Height) // túlment az ütőn
                {
                    _BallNextPosition.Top = 0;
                    StopGame(false);
                }
            }
        }

        /// <summary>
        /// Labda animálása.
        /// </summary>
        private void AnimateBall()
        {
            _BallAnimation = new ThicknessAnimation(); // új animáció létrehozása
            _BallAnimation.From = _BallCurrentPosition; // a labda jelenlegi pozíciójából
            _BallAnimation.To = _BallNextPosition; // az új pozícióba
            _BallAnimation.Duration = new Duration(TimeSpan.FromMilliseconds(5)); // animáció ideje
            _BallAnimation.SpeedRatio = _Speed / BallTravelDistance(); // az animáció sebessége függ a játék sebességétől, illetve a labda utazási távolságától

            _EllipseBall.BeginAnimation(Rectangle.MarginProperty, _BallAnimation, HandoffBehavior.SnapshotAndReplace);
            // animáció lejátszása az adott objektumon és tulajdonságon
            // ha már fut animáció, akkor annak leállítása
        }

        /// <summary>
        /// Labda utazási távolságának kiszámítása.
        /// </summary>
        /// <returns></returns>
        private Double BallTravelDistance()
        {
            return (Math.Sqrt(Math.Pow(Math.Abs(_BallNextPosition.Left - _BallCurrentPosition.Left), 2) + Math.Pow(Math.Abs(_BallNextPosition.Top - _BallCurrentPosition.Top), 2))) * 2;
        }

        /// <summary>
        /// Saját ütő animálása.
        /// </summary>
        /// <param name="x">A mozgatás mértéke pontban.</param>
        private void AnimatePlayerPad(Int32 x)
        {
            _PlayerPadAnimation = new ThicknessAnimation();
            _PlayerPadAnimation.From = _PlayerPadCurrentPosition;
            _PlayerPadAnimation.To = new Thickness(_RectanglePlayerPad.Margin.Left + x, _RectanglePlayerPad.Margin.Top, _RectanglePlayerPad.Margin.Right - x, _RectanglePlayerPad.Margin.Bottom);
            _PlayerPadAnimation.Duration = new Duration(TimeSpan.FromMilliseconds(100));

            _RectanglePlayerPad.BeginAnimation(Rectangle.MarginProperty, _PlayerPadAnimation, HandoffBehavior.SnapshotAndReplace);
        }

        /// <summary>
        /// Ellenfél ütőjének animálása.
        /// </summary>
        /// <param name="x">A mozgatás mértéke pontban.</param>
        private void AnimateOpponentPad(Int32 x)
        {
            _OpponentPadAnimation = new ThicknessAnimation();
            _OpponentPadAnimation.From = _OpponentPadCurrentPosition;
            _OpponentPadAnimation.To = new Thickness(_RectangleOpponentPad.Margin.Left + x, _RectangleOpponentPad.Margin.Top, _RectangleOpponentPad.Margin.Right - x, _RectangleOpponentPad.Margin.Bottom);
            _OpponentPadAnimation.Duration = new Duration(TimeSpan.FromMilliseconds(100));

            _RectangleOpponentPad.BeginAnimation(Rectangle.MarginProperty, _OpponentPadAnimation, HandoffBehavior.SnapshotAndReplace);
        }

        private void AnimateGameStartLabel()
        {
            _GameStartLabelOpacityAnimation = new DoubleAnimation(); // áttetszőség animálása
            _GameStartLabelOpacityAnimation.From = 1;
            _GameStartLabelOpacityAnimation.To = 0;
            _GameStartLabelOpacityAnimation.Duration = new Duration(TimeSpan.FromSeconds(1));

            _GameStartLabelFontSizeAnimation = new DoubleAnimation(); // méret animálása
            _GameStartLabelFontSizeAnimation.From = 24;
            _GameStartLabelFontSizeAnimation.To = 100;
            _GameStartLabelFontSizeAnimation.Duration = new Duration(TimeSpan.FromSeconds(1));

            // mindkét animációt elindítjuk, csak más tulajdonságon:
            _LabelGameStart.BeginAnimation(Label.FontSizeProperty, _GameStartLabelFontSizeAnimation, HandoffBehavior.SnapshotAndReplace);
            _LabelGameStart.BeginAnimation(Label.OpacityProperty, _GameStartLabelOpacityAnimation, HandoffBehavior.SnapshotAndReplace);
        }
    }
}
