﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using ELTE.Net.GameServer.Shared;
using System.Diagnostics;

namespace ELTE.Net.GameServer.Service
{
    /// <summary>
    /// Szerverkezelő típusa.
    /// </summary>
    class ServerManager : IDisposable
    {
        private const Int32 _ServerPort = 4503; // a fogadás portja

        private static ServerManager _Instance; // egyke objektum

        /// <summary>
        /// Szerverkezelő példányának lekérdezése.
        /// </summary>
        public static ServerManager Instance 
        { 
            get
            {
                if (_Instance == null)
                    _Instance = new ServerManager(); // lusta betöltést alkalmazunk
                return _Instance;                 
            }         
        }

        private Socket _ServerSocket; // szerver fogadó csatolója
        private Boolean _Running; // fut-e a szerver

        private List<Game> _OnlineGames; // játékok listája
        private List<ClientManager> _Clients; // kliensek listája

        /// <summary>
        /// Futási állapot lekérdezése.
        /// </summary>
        public Boolean Running
        {
            get { return _ServerSocket != null && _ServerSocket.IsBound; }
        }

        /// <summary>
        /// Szerverkezelő példányosítása.
        /// </summary>
        private ServerManager()
        {
            _Clients = new List<ClientManager>();
            _OnlineGames = new List<Game>();
        }

        /// <summary>
        /// Szerver indítása.
        /// </summary>
        public void Start()
        {
            Thread serverThread = new Thread(WatchNewConnections); // szál az új kapcsolatoknak
            serverThread.Start();

            Thread watchThread = new Thread(WatchCurrentConnections); // szál a jelenlegi kapcsolatok figyelésére
            watchThread.Start();
        }
        /// <summary>
        /// Szerver leállítása.
        /// </summary>
        public void Stop()
        {
            if (_Running)
            {
                _Running = false;
                foreach (ClientManager client in _Clients)
                    client.Disconnect(); // minden klienst leválasztunk

                if (_ServerSocket != null && _ServerSocket.IsBound) // lekapcsoljuk a szerver csatolót
                    _ServerSocket.Close();
            }
        }
        /// <summary>
        /// Nyitott játékok lekérdezése.
        /// </summary>
        /// <param name="player"></param>
        /// <returns></returns>
        public Game[] GetOpenGames(Player player)
        {
            return _OnlineGames.Where(x => x.Phase == GamePhase.Open && !x.FirstPlayer.Equals(player)).ToArray();
        }
        /// <summary>
        /// Új játék létrehozása.
        /// </summary>
        /// <param name="game">Játék.</param>
        public void CreateGame(Game game)
        {
            game.GameId = Guid.NewGuid().GetHashCode(); // azonosító hozzárendelése
            game.CreateTime = DateTime.Now;
            _OnlineGames.Add(game);
        }

        /// <summary>
        /// Csatalkozás nyitott játékhoz.
        /// </summary>
        /// <param name="gameId">Játék azonosítója.</param>
        /// <param name="player">Játékos.</param>
        /// <returns>Sikeres volt-e a csatlakozás.</returns>
        public Boolean JoinGame(Int32 gameId, Player player)
        {
            Game game = GetGame(gameId); // keresés a játékra
            if (game != null)
            {
                lock (game) // kölcsönös kizárás a játékra
                {
                    if (game.Phase == GamePhase.Open) // amennyiben nyitott a játék
                    {
                        game.SecondPlayer = player; // elindítjuk
                        game.StartTime = DateTime.Now;
                        MessagePlayer(game.FirstPlayer, MessageCode.JoinGame, game); // üzenünki a másik játékosnak, hogy kezdődhet a játék
                    }
                    return true;
                }
            }
            else
                return false;

        }
        /// <summary>
        /// Játék lekérdezése.
        /// </summary>
        /// <param name="gameId">Játék azonosítója.</param>
        /// <returns>A játék, vagy null, ha nem sikerült megtalálni.</returns>
        public Game GetGame(Int32 gameId)
        {
            return _OnlineGames.Where(x => x.GameId == gameId).FirstOrDefault();
        }
        /// <summary>
        /// Játékállapot változtatás.
        /// </summary>
        /// <param name="player">Játékos.</param>
        /// <param name="game">Játék.</param>
        /// <param name="value">A változtatás értéke.</param>
        public void ChangeGameState(Player player, Game game, Int32 value)
        {
            if (game.Phase == GamePhase.Playing) // amennyiben fut a játék
            {
                ClientManager client = _Clients.Where(x => x.Player == ((game.FirstPlayer == player) ? game.SecondPlayer : game.FirstPlayer)).FirstOrDefault();
                // lekérdezzük a klienst
                if (client != null)
                {
                    client.SendMessage(MessageCode.ChangeGameState, value); // továbbküljük neki az állapotváltozást
                }
            }
        }
        /// <summary>
        /// Játék befejezése.
        /// </summary>
        /// <param name="game">Játék.</param>
        /// <param name="player">A befejező játékos.</param>
        public void EndGame(Game game, Player player)
        { 
            lock(game) // kölcsönös kizárás a játékra
            {
                if (game.Phase != GamePhase.Completed) // ha még nincs vége a játéknak
                {
                    if (player == game.FirstPlayer) // jelezünk a másik játékosnak
                        MessagePlayer(game.SecondPlayer, MessageCode.EndGame);
                    else
                        MessagePlayer(game.FirstPlayer, MessageCode.EndGame);

                    game.EndTime = DateTime.Now;
                    _OnlineGames.Remove(game); // töröljük a játékot
                }
            }
        }

        /// <summary>
        /// Példány megsemmisítése.
        /// </summary>
        public void Dispose()
        {
            Stop();
        }

        /// <summary>
        /// Kapcsolatok fogadása.
        /// </summary>
        private void WatchNewConnections()
        {
            try
            {
                String serverName = Dns.GetHostName(); // a szerver DNS nevének lekérdezése
                List<IPAddress> addressList = new List<IPAddress>();
                IPHostEntry ipEntry = Dns.GetHostEntry(serverName); // a szerver nevének megfelelő címek lekérdezése
                foreach (IPAddress address in ipEntry.AddressList)
                {
                    if (address.AddressFamily == AddressFamily.InterNetwork) // ha ez IPv4 cím
                    {
                        addressList.Add(address); // akkor felvesszük a lehetséges címek közé
                    }
                }

                if (addressList.Count >= 1) // nem sikerült címet kiolvasni
                {
                    _Running = true;
                    _ServerSocket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
                    _ServerSocket.Bind(new IPEndPoint(addressList[0], _ServerPort)); // csatolás IP címhez
                    _ServerSocket.Listen(10); // maximum 10 kapcsolat

                    while (_Running)
                    {
                        Socket clientSocket = _ServerSocket.Accept(); // elfogadjuk a csatlakozást
                        _Clients.Add(new ClientManager(clientSocket)); // felvesszük a klienst
                    }
                }
            }
            catch
            {
                Stop();
            }
        }

        /// <summary>
        /// Szerver felügyelete.
        /// </summary>
        private void WatchCurrentConnections()
        {
            while (_Running) // ha fut a szerver
            {
                Thread.Sleep(10000); // 10 másodperces késleltetés
                for (Int32 i = _Clients.Count - 1; i >= 0; i--)
                {
                    if (!_Clients[i].Connected) // ha már nincs csatlakozva a kliens
                    {
                        _Clients.RemoveAt(i); // akkor kitöröljük a listából
                    }
                }
            }
        }

        /// <summary>
        /// Üzenet küldése adott játékosnak.
        /// </summary>
        /// <param name="player">Játékos.</param>
        /// <param name="messageCode">Üzenet kódja.</param>
        /// <param name="messageContent">Üzenet tartalma.</param>
        public void MessagePlayer(Player player, MessageCode messageCode, Object messageContent = null)
        {
            ClientManager clientManager = _Clients.Where(x => x.Player.Equals(player)).FirstOrDefault(); // keresés a játékos alapján
            if (clientManager != null)
                clientManager.SendMessage(messageCode, messageContent);
        }
    }
}
