﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.Serialization;
using System.ServiceModel;
using System.Text;

namespace WCFSample.Service
{
    /// <summary>
    /// A szolgáltatás implementációja.
    /// </summary>
    [ServiceBehavior(InstanceContextMode=InstanceContextMode.PerSession)]
    public class SampleService : ISampleService
    {
        /// <summary>
        /// A tárolt elemek.
        /// </summary>
        private static readonly List<Item> _items = new List<Item>();

        /// <summary>
        /// Feliratkozott kliensek.
        /// </summary>
        private static readonly List<ISampleCallback> _clients = new List<ISampleCallback>();

        /// <summary>
        /// A következő kliensnek kiosztandó azonosító.
        /// </summary>
        private static UInt32 _nextId = 0;

        /// <summary>
        /// Kliens azonosítója.
        /// </summary>
        private UInt32 _id;

        /// <summary>
        /// Konstruktor.
        /// Azonosító kiosztása a felcsatlakozott kliensnek.
        /// </summary>
        public SampleService()
        {
            _id = _nextId++;
        }

        /// <summary>
        /// Új elem felvétele.
        /// </summary>
        /// <param name="item">Az új elem.</param>
        public void AddItem(Item item)
        {
            LogEvent("Elem felévtele.");
            _items.Add(item);

            _clients.ForEach(delegate(ISampleCallback callback)
            {
                ICommunicationObject comm = (ICommunicationObject)callback;
                if (comm.State == CommunicationState.Opened)
                    callback.OnItemAdded(item);
                else if (comm.State == CommunicationState.Faulted ||
                    comm.State == CommunicationState.Closing ||
                    comm.State == CommunicationState.Closed)
                {
                    if(comm.State == CommunicationState.Faulted)
                        comm.Abort();

                    _clients.Remove(callback);
                    LogEvent("Kapcsolat megszakadt.");
                }
            });
        }

        /// <summary>
        /// Elemek lekérdezése.
        /// </summary>
        /// <returns>A tárolt elemek.</returns>
        public Item[] GetItems()
        {
            LogEvent("Elemek lekérdezése.");
            return _items.ToArray();
        }

        /// <summary>
        /// Feliratkozás az eseményekről való értesítésekre.
        /// </summary>
        public void Subscribe()
        {
            ISampleCallback callback = OperationContext.Current.GetCallbackChannel<ISampleCallback>();
            if (!_clients.Contains(callback))
            {
                _clients.Add(callback);
                LogEvent("Feliratkozás.");
            }
        }

        /// <summary>
        /// Leiratkozás az eseményekről való értesítésekről.
        /// </summary>
        public void Unsubscribe()
        {
            ISampleCallback callback = OperationContext.Current.GetCallbackChannel<ISampleCallback>();
            if (_clients.Contains(callback))
            {
                _clients.Remove(callback);
                LogEvent("Leiratkozás.");
            }
        }

        /// <summary>
        /// Esemény naplózása.
        /// </summary>
        /// <param name="message">Üzenet.</param>
        private void LogEvent(String message)
        {
            Console.WriteLine(String.Format("[Id: {1} | Time: {2}]  {0}", message, _id, DateTime.Now.ToLongTimeString()));
        }
    }
}
